/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "cg_local.h"

int cg_numSolids;
static entity_state_t *cg_solidList[MAX_PARSE_ENTITIES];

int cg_numTriggers;
static entity_state_t *cg_triggersList[MAX_PARSE_ENTITIES];
static qboolean		cg_triggersListTriggered[MAX_PARSE_ENTITIES];

//===================
//CG_PredictionActive
//===================
qboolean CG_PredictionActive( void )
{
	if( cgs.demoPlaying || (cg.frame.playerState.pmove.pm_flags & PMF_NO_PREDICTION) )
		return qfalse;

	return (cg_predict->integer);
}

//===================
//CG_WeaponPredictionActive
//===================
#ifdef PREDICTSHOOTING
qboolean CG_WeaponPredictionActive( void )
{
	if( !CG_PredictionActive() )
		return qfalse;

	if( cg.frame.playerState.pmove.pm_type != PM_NORMAL )
		return qfalse;

	return (cg_predictWeapon->integer);
}
#endif

//===================
//CG_CheckPredictionError
//===================
void CG_CheckPredictionError( void )
{
	int		delta[3];
	int		frame;

	if( !CG_PredictionActive() )
		return;

	// calculate the last usercmd_t we sent that the server has processed
	frame = cg.frame.ucmdExecuted & CMD_MASK;

	// compare what the server returned with what we had predicted it to be
	VectorSubtract( cg.frame.playerState.pmove.origin, cg.predictedOrigins[frame], delta );

	// save the prediction error for interpolation
	if( abs(delta[0]) > 128 || abs(delta[1]) > 128 || abs(delta[2]) > 128 ) {
		if( cg_showMiss->integer )
			CG_Printf( "prediction miss on %i: %i\n", cg.frame.serverFrame, abs(delta[0]) + abs(delta[1]) + abs(delta[2]) );
		VectorClear( cg.predictionError );					// a teleport or something
	} else {
		if( cg_showMiss->integer && (delta[0] || delta[1] || delta[2]) )
			CG_Printf( "prediction miss on %i: %i\n", cg.frame.serverFrame, abs(delta[0]) + abs(delta[1]) + abs(delta[2]) );
		VectorCopy( cg.frame.playerState.pmove.origin, cg.predictedOrigins[frame] );
		VectorCopy( delta, cg.predictionError );	// save for error interpolation
	}
}

//====================
//CG_BuildSolidList
//====================
void CG_BuildSolidList( void )
{
	int	i;

	entity_state_t *ent;

	cg_numSolids = 0;
	cg_numTriggers = 0;
	for( i = 0; i < cg.frame.numEntities; i++ ) {
		ent = &cg.frame.parsedEntities[i & (MAX_PARSE_ENTITIES-1)];
		if( ent->type == ET_ITEM || //items are always triggers, but don't have a brush model
			(ent->solid == SOLID_BMODEL && ent->type == ET_PUSH_TRIGGER) ) {
			cg_triggersList[cg_numTriggers++] = ent;
		}
		else if( ent->solid ) {
			cg_solidList[cg_numSolids++] = ent;
		}
	}
}

//====================
//CG_ClipEntityContact
//====================
static qboolean CG_ClipEntityContact( vec3_t origin, vec3_t mins, vec3_t maxs, int entNum, qboolean interpolated )
{
	centity_t		*cent;
	struct cmodel_s	*cmodel;
	trace_t			tr;
	vec3_t			absmins, absmaxs;
	vec3_t			entorigin, entangles;

	if( !mins )
		mins = vec3_origin;
	if( !maxs )
		maxs = vec3_origin;

	// find the cmodel
	cmodel = CG_CModelForEntity( entNum );
	if( !cmodel )
		return qfalse;

	cent = &cg_entities[entNum];

	// find the origin
	if( cent->current.solid == SOLID_BMODEL ) {	// special value for bmodel
		if( interpolated )
		{
			int j;
			VectorLerp( cent->prev.origin, cg.lerpfrac, cent->current.origin, entorigin );
			for( j = 0; j < 3; j++ )
				entangles[j] = LerpAngle(cent->prev.angles[j], cent->current.angles[j], cg.lerpfrac );
		}
		else {
			VectorCopy( cent->current.origin, entorigin );
			VectorCopy( cent->current.angles, entangles );
		}
	} else {							// encoded bbox
		if( interpolated ) {
			VectorLerp( cent->prev.origin, cg.lerpfrac, cent->current.origin, entorigin );
			VectorClear( entangles );	// boxes don't rotate
		} else {
			VectorCopy( cent->current.origin, entorigin );
			VectorClear( entangles );	// boxes don't rotate
		}
	}

	// convert the box to compare to absolute coordinates
	VectorAdd( origin, mins, absmins );
	VectorAdd( origin, maxs, absmaxs );
	trap_CM_TransformedBoxTrace( &tr, vec3_origin, vec3_origin, absmins, absmaxs, cmodel, MASK_ALL, entorigin, entangles );
	return tr.startsolid || tr.allsolid;
}

//====================
//CG_Predict_TouchTriggers
//====================
static void CG_Predict_TouchTriggers( pmove_t *pm ) {
	int			i;
	entity_state_t	*state;

	// fixme: more accurate check for being able to touch or not
	if( pm->s.pm_type != PM_NORMAL ) {
		return;
	}

	for( i = 0; i < cg_numTriggers; i++ ) {
		state = cg_triggersList[i];

		if( state->type == ET_PUSH_TRIGGER ) {
			if( !cg_triggersListTriggered[i] ) {
				if( CG_ClipEntityContact( cg.predictedOrigin, pm->mins, pm->maxs, state->number, qfalse ) ) {
					GS_TouchPushTrigger( &pm->s, state );
					cg_triggersListTriggered[i] = qtrue;
				}
			}
		}
	}
}

//====================
//CG_ClipMoveToEntities
//====================
static void CG_ClipMoveToEntities( vec3_t start, vec3_t mins, vec3_t maxs, vec3_t end, int ignore, int contentmask, trace_t *tr )
{
	int			i, x, zd, zu;
	trace_t		trace;
	vec3_t		origin, angles;
	entity_state_t	*ent;
	struct cmodel_s	*cmodel;
	vec3_t		bmins, bmaxs;

	for( i = 0; i < cg_numSolids; i++ ) {
		ent = cg_solidList[i];

		if( ent->number == ignore )
			continue;
		if( !(contentmask & CONTENTS_CORPSE) && ((ent->type == ET_CORPSE)||(ent->type == ET_GIB)) )
			continue;

		if ( ent->solid == SOLID_BMODEL ) {	// special value for bmodel
			cmodel = trap_CM_InlineModel( ent->modelindex );
			if( !cmodel )
				continue;

			VectorCopy( ent->origin, origin );
			VectorCopy( ent->angles, angles );
		} else {							// encoded bbox
			x = 8 * (ent->solid & 31);
			zd = 8 * ((ent->solid>>5) & 31);
			zu = 8 * ((ent->solid>>10) & 63) - 32;

			bmins[0] = bmins[1] = -x;
			bmaxs[0] = bmaxs[1] = x;
			bmins[2] = -zd;
			bmaxs[2] = zu;

			VectorCopy( ent->origin, origin );
			VectorClear( angles );	// boxes don't rotate
			cmodel = trap_CM_ModelForBBox( bmins, bmaxs );
		}

		trap_CM_TransformedBoxTrace( &trace, start, end, mins, maxs, cmodel, contentmask, origin, angles );
		if( trace.allsolid || trace.fraction < tr->fraction ) {
			trace.ent = ent->number;
			*tr = trace;
		} else if( trace.startsolid ) {
			tr->startsolid = qtrue;
		}
		if( tr->allsolid )
			return;
	}
}

//================
//CG_Trace
//================
void CG_Trace( trace_t *t, vec3_t start, vec3_t mins, vec3_t maxs, vec3_t end, int ignore, int contentmask )
{
	// check against world
	trap_CM_BoxTrace( t, start, end, mins, maxs, NULL, contentmask );
	t->ent = t->fraction < 1.0 ? 0 : -1; // world entity is 0
	if( t->fraction == 0 )
		return;			// blocked by the world

	// check all other solid models
	CG_ClipMoveToEntities( start, mins, maxs, end, ignore, contentmask, t );
}

//================
//CG_PointContents
//================
int	CG_PointContents( vec3_t point )
{
	int			i;
	entity_state_t	*ent;
	struct cmodel_s	*cmodel;
	int			contents;

	contents = trap_CM_PointContents( point, NULL );

	for( i = 0; i < cg_numSolids; i++ ) {
		ent = cg_solidList[i];
		if( ent->solid != SOLID_BMODEL )	// special value for bmodel
			continue;

		cmodel = trap_CM_InlineModel( ent->modelindex );
		if( cmodel )
			contents |= trap_CM_TransformedPointContents ( point, cmodel, ent->origin, ent->angles );
	}

	return contents;
}


float	predictedSteps[CMD_BACKUP];	// for step smoothing
//=================
//CG_PredictAddStep
//=================
static void CG_PredictAddStep( int virtualtime, int predictiontime, float stepSize ) {
	
	float	oldStep;
	int		delta;

	// check for stepping up before a previous step is completed
	delta = cg.realTime - cg.predictedStepTime;
	if( delta < PREDICTED_STEP_TIME ) {
		oldStep = cg.predictedStep * ((float)(PREDICTED_STEP_TIME - delta) / (float)PREDICTED_STEP_TIME);
	} else {
		oldStep = 0;
	}

	cg.predictedStep = oldStep + stepSize;
	cg.predictedStepTime = cg.realTime - ( predictiontime - virtualtime );
}

//=================
//CG_PredictSmoothSteps
//=================
static void CG_PredictSmoothSteps( void ) {
	int			outgoing;
	int			frame;
	usercmd_t	cmd;
	int			i;
	int			virtualtime = 0, predictiontime = 0;

	cg.predictedStepTime = 0;
	cg.predictedStep = 0;
	
	trap_NET_GetCurrentState( NULL, &outgoing, NULL );

	i = outgoing;
	while( predictiontime < PREDICTED_STEP_TIME )
	{
		if( outgoing - i >= CMD_BACKUP )
			break;
		
		frame = i & CMD_MASK;
		trap_NET_GetUserCmd( frame, &cmd );
		predictiontime += cmd.msec;
		i--;
	}

	// run frames
	while( ++i <= outgoing ) {
		frame = i & CMD_MASK;
		trap_NET_GetUserCmd( frame, &cmd );
		virtualtime += cmd.msec;

		if( predictedSteps[frame] ) {
			CG_PredictAddStep( virtualtime, predictiontime, predictedSteps[frame] );
		}
	}
}

//=================
//CG_PredictMovement
//
//Sets cg.predictedVelocty, cg.predictedOrigin and cg.predictedAngles
//=================
void CG_PredictMovement( void )
{
	int				executed, outgoing;
	int				frame;
	usercmd_t		cmd;
	pmove_t			pm;
	int				i;

	cg.predictedTime = 0;

	if( !CG_PredictionActive() ) {
		// just set angles
		trap_NET_GetUserCmd( trap_NET_GetCurrentUserCmdNum(), &cmd ); // the command being built but yet unsent
		VectorCopy( cg.frame.playerState.pmove.velocity, cg.predictedVelocity );
		VectorCopy( cg.frame.playerState.pmove.origin, cg.predictedOrigin );
		for( i = 0; i < 3; i++ )
			cg.predictedAngles[i] = SHORT2ANGLE( cmd.angles[i] ) + SHORT2ANGLE( cg.frame.playerState.pmove.delta_angles[i] );
		return;
	}

	trap_NET_GetCurrentState( NULL, &outgoing, NULL );
	executed = cg.frame.ucmdExecuted;

	// if we are too far out of date, just freeze
	if( outgoing - executed >= CMD_BACKUP ) {
		if( cg_showMiss->integer )
			CG_Printf( "exceeded CMD_BACKUP\n" );
		return;
	}
	
	// copy current state to pmove
	memset( &pm, 0, sizeof(pm) );
	pm.s = cg.frame.playerState.pmove;
	pm.passent = cg.frame.playerState.POVnum;
	if( cg.frame.playerState.stats[STAT_GAMETYPE] == GAMETYPE_RACE )
		pm.contentmask = MASK_DEADSOLID;
	else
		pm.contentmask = MASK_PLAYERSOLID;

	// wsw : jal : set max_walljump count for gametype
	pm.max_walljumps = GS_GameType_MaxWallJumps( cg.frame.playerState.stats[STAT_GAMETYPE] );

	// clear the triggered toggles for this prediction round
	memset( &cg_triggersListTriggered, qfalse, sizeof(cg_triggersListTriggered) );

	// run frames
	while( ++executed <= outgoing ) {
		frame = executed & CMD_MASK;
		trap_NET_GetUserCmd( frame, &pm.cmd );
		cg.predictedTime += pm.cmd.msec;
		Pmove( &pm );

		// update the prediction values
		cg.groundEntity = pm.groundentity;

		VectorCopy( pm.s.velocity, cg.predictedVelocity );
		VectorCopy( pm.s.origin, cg.predictedOrigin );
		VectorCopy( pm.viewangles, cg.predictedAngles );

		// Check for touching triggers
		CG_Predict_TouchTriggers( &pm );

		//copy for stair smoothing
		predictedSteps[frame] = pm.step;

		// save for debug checking
		VectorCopy( pm.s.origin, cg.predictedOrigins[frame] ); // store for prediction error checks
	}

	CG_PredictSmoothSteps();
}

//=================
//CG_PredictWeapon_Check
//
// Check whether we managed to predict correctly
//=================
#ifdef PREDICTSHOOTING
static void CG_PredictWeapon_Check( int command, const weapon_state_t *state, qboolean check )
{
	predicted_weapon_state_t *predicted, *oldest;
	int i;

	oldest = &cg.predictedWeaponStates[0];
	for( i = 0, predicted = &cg.predictedWeaponStates[0]; i < CMD_BACKUP; i++, predicted++ )
	{
		if( predicted->command == command )
		{
			if( !check )
				break;

			if( predicted->state.nexttime != state->nexttime ) {
				Com_Printf( "Weapon prediction miss, command %i: Wrong nexttime, predicted %i, was %i\n", command,
					predicted->state.nexttime, state->nexttime );
			}
			if( predicted->state.status != state->status ) {
				Com_Printf( "Weapon prediction miss, command %i: Wrong status, predicted %i, was %i\n", command,
					predicted->state.status, state->status );
			}
			if( predicted->state.changing != state->changing ) {
				Com_Printf( "Weapon prediction miss, command %i: Wrong changing, predicted %i, was %i\n", command,
					predicted->state.changing, state->changing );
			}
			break;
		}
		if( predicted->command < oldest->command ) {
			oldest = predicted;
		}
	}
	if( i == CMD_BACKUP )
		predicted = oldest;

	predicted->command = command;
	predicted->state = *state;
}
#endif

//=================
//CG_PredictWeapon
//
//Sets cg.predictedWeapon, cg.predictedWeaponEvent and cg.predictedWeaponStatus
//=================
#ifdef PREDICTSHOOTING
void CG_PredictWeapon( void )
{
	int				executed, sent;
	int				frame, i;
	pmove_t			pm;
	weapon_state_t	weaponstate;
	int				latched_weapon;

	cg.predictedWeapon = cg.frame.playerState.stats[STAT_WEAPON_ITEM];
	cg.predictedWeaponEvent = 0;
	cg.predictedWeaponStatus = WEAPON_READY;

	if( !CG_WeaponPredictionActive() )
		return;

	trap_NET_GetCurrentState( NULL, NULL, &sent );
	executed = cg.frame.ucmdExecuted;

	// if we are too far out of date, just freeze
	if( sent - executed >= CMD_BACKUP ) {
		Com_Printf( "PredictWeapon: exceeded CMD_BACKUP\n" );
		return;
	}

	// copy current state to weaponstate
	memset( &weaponstate, 0, sizeof(weaponstate) );
	if( cg.frame.playerState.stats[STAT_WEAPON_STATE_STATUS] & STAT_WEAPON_STATE_STATUS_CHANGING ) {
		weaponstate.changing = qtrue;
		weaponstate.status = cg.frame.playerState.stats[STAT_WEAPON_STATE_STATUS];
		weaponstate.status &= ~STAT_WEAPON_STATE_STATUS_CHANGING;
	} else {
		weaponstate.changing = qfalse;
		weaponstate.status = cg.frame.playerState.stats[STAT_WEAPON_STATE_STATUS];
	}
	weaponstate.nexttime = cg.frame.playerState.stats[STAT_WEAPON_STATE_NEXTTIME];
	// TODO: powered time

	// see what we were changing to
	latched_weapon = -1;
	if( weaponstate.changing ) {
		i = (cg.weaponChangesPosition + 1) % CMD_BACKUP;
		do {
			if( cg.weaponChanges[i][0] && cg.weaponChanges[i][0] <= executed )
				latched_weapon = cg.weaponChanges[i][1];
			i = (i + 1) % CMD_BACKUP;
		} while( i != (cg.weaponChangesPosition + 1) % CMD_BACKUP );

		if( latched_weapon == -1 ) {
			Com_Printf( "Warning: Unpredicted weapon change\n" );
			latched_weapon = cg.predictedWeapon;
		}
	}

	if( cg_showMiss->integer )
		CG_PredictWeapon_Check( executed, &weaponstate, qtrue );

	// run frames
	// we don't predict the latest ucmd, because we only want to predict it once, and it isn't complete
	while( ++executed < sent ) {
		frame = executed & CMD_MASK;
		trap_NET_GetUserCmd( frame, &pm.cmd );

		// check if we started changing weapon at this point
		i = (cg.weaponChangesPosition + 1) % CMD_BACKUP;
		do {
			if( cg.weaponChanges[i][0] == executed ) {
				latched_weapon = cg.weaponChanges[i][1];
				weaponstate.changing = qtrue;
				break;
			} else if( cg.weaponChanges[i][0] > executed ) {
				break;
			}
			i = (i + 1) % CMD_BACKUP;
		} while( i != (cg.weaponChangesPosition + 1) % CMD_BACKUP );

		CG_Predict_Weaponstate( &cg.predictedWeapon, &latched_weapon, &weaponstate, executed, &pm.cmd,
			(executed > cg.predictedWeaponCommand) );
		if( cg_showMiss->integer )
			CG_PredictWeapon_Check( executed, &weaponstate, (cg_showMiss->integer == 2) );
	}

	cg.predictedWeaponStatus = weaponstate.status;
	cg.predictedWeaponCommand = executed-1;

	CG_vWeapUpdateState();
}
#endif
